
// #define _XOPEN_SOURCE 600

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include "http.h"
#include "utils.h"
#include "soket.h"

void http_set_content_type(char *file, char *content_type)
{
	char *suffix;

	if (!file || !content_type)
		return;

	suffix = rindex(file, '.');
	if ((suffix) /*a la 'file'? */&& (strlen(suffix)>1)) {

		suffix++;

		if (strstr(suffix, "txt") && (strlen(suffix) == 3))
			strcpy(content_type, "text/plain; charset=utf-8");

		else if (strstr(suffix, "html") && (strlen(suffix) == 4) )
			strcpy(content_type, "text/html; charset=utf-8");

		else if (strstr(suffix, "htm") && (strlen(suffix) == 3) )
			strcpy(content_type, "text/html; charset=utf-8");

		else if (strstr(suffix, "png") && (strlen(suffix) == 3))
			strcpy(content_type, "image/png");

		else if (strstr(suffix, "jpg") && (strlen(suffix) == 3))
			strcpy(content_type, "image/jpeg");

		else if (strstr(suffix, "jpeg") && (strlen(suffix) == 4))
			strcpy(content_type, "image/jpeg");

		else if (strstr(suffix, "css") && (strlen(suffix) == 3))
			strcpy(content_type, "text/css" );

		else if (strstr(suffix, "mp3") && (strlen(suffix) == 3))
			strcpy(content_type, "audio/mpeg");

		/* Add more content types here */

		else
			strcpy(content_type, "application/octet-stream");
	} else {
		strcpy(content_type, "application/octet-stream");
	}
}

int http_bad_method(char *m)
{
/*
	HTTP methods:

	OPTIONS
	GET
	HEAD
	POST
	PUT
	DELETE
	TRACE
	CONNECT
*/

	char *get = "GET";
	char *head = "HEAD";

	/* Dynamic Content Generator Extension */
	char *dcgp = "DCGP";

//	char *post = "POST"; // hard work )

	if ((strstr(m, get)) && (strlen(m) == strlen(get)))
		return 0;

	if ((strstr(m, head)) && (strlen(m) == strlen(head)))
		return 0;

	if ((strstr(m, dcgp)) && (strlen(m) == strlen(dcgp)))
		return 0;

	return 1;
}

/*
 * Returns:
 * 0 - 'field' dos not match 'name'
 * 1 - OK
 * 2 - already set
 */
int http_set_param(char *field, char *val, char *name, char **ptr)
{
	char *s;

	// printf("fld = %s, val = %s, name = %s\n", field, val, name);

	s = strstr(field, name);
	if ((!s) || (strlen(field) != strlen(name)))
		return 0;

	if (*ptr)
		return 2;

	*ptr = val;
	return 1;
}

// the function is too long, split it ...
int http_parse_request_header(char *data, struct http_request_header *h)
{
	int r;
	int ver, rev;
	char *s, *tok, *l;
	int res;

//	printf("\n### Request:\n%s\n", data);

	s = strstr(data, "\r\n");
	if (!s)
		return -400;

	if ((s - data) >= MAX_REQUEST_LINE_LEN - 1) {
		strncpy(h->request_line, data, MAX_REQUEST_LINE_LEN - 1);
		return -413;
	}

	strncpy(h->request_line, data, s - data);

	/* dealing with method (leading spaces already skipped) */
	h->method = tok = data;
	while (!isspace(tok[0]) && ((tok-data) < MAX_METH_LEN))
		tok++;
	if ((tok-data) >= MAX_METH_LEN)
		return -400;

	tok[0] = 0;
	str_to_upper_case(h->method);
	if (http_bad_method(h->method))
		return -501;

	/* and with uri */
	tok++;
	while (isspace(tok[0]))
		tok++;

	/* TODO
	 * all HTTP/1.1 servers MUST accept the absoluteURI form in requests
	 * even though HTTP/1.1 clients will only generate them in requests to proxies
	 */

	/* skip over 'http://host.domain.dom' if present -->>> swebs.c/swebs_parse_uri() */
	/*
	scheme = strstr(tok, "http://"); but mind case...
	if (scheme) {
		if (tok != scheme)
			return -400;
		tok += strlen("http://");
		while (tok[0] != '/')
			tok++;
	}
	*/

	if (tok[0] == '/')
		tok++;
	h->file = tok;

	while (!isspace(tok[0]))
		tok++;
	tok[0] = 0;

	l = index(h->file, '?');
	if (l) {
		l[0] = 0;
		h->args = l + 1;
	};

	/* and with HTTP/v.s */
	tok++;
	while (isspace(tok[0]))
		tok++;
	h->http = tok;
	while ((!isspace(tok[0])) && ((tok - h->http) < MAX_HTTP_LEN))
		tok++;
	tok[0] = 0;

	/* do we have some gift between 'HTTP/x.x' and '\r\n'? */
	if (s - tok) {
		printf("ALARM! :). %d bytes between 'HTTP/x.x' and '\r\n'\n", s - tok);
		/* and put the gift somewhere */
		return -402; /* :) */
	}

	r = sscanf(h->http, "HTTP/%d.%d", &ver, &rev);
	if (r != 2)
		return -400;

	ver = rev = 1; //:)
	if ((ver != 1) || (rev != 1))
		return -505;

	// make it also separate func... (on http layer!)?
	do {
		char *colon;
		char *prm_name, *prm_val;

		/* "\r\n" -> 0*/
		s[0] = s[1] = 0;
		s += 2;

		/* folding is unimplemented */
		if (isblank(s[0]))
			return -501;
		/* or replace all '\r\nSP|HT' -> '   ' */

		l = strstr(s, "\r\n");
		if (!l)
			return -400;

		if (l == s) /* end of header */
			break;

		colon = index(s, ':');
		if (!colon)
			return -400;
		/* no spaces between 'field' and ':'! */
		if (isblank(colon[-1]))
			return -400;
		colon[0] = 0;
		prm_name = s;
		str_to_upper_case(prm_name);
		// printf("prm_name = %s\n", prm_name);

		prm_val = colon + 1;
		while (isblank(prm_val[0]))
			prm_val++;

		/* no empty fields! */
		if (prm_val[0] == '\r')
			return -400;
		l[0] = 0;

		do {
			res = http_set_param(prm_name, prm_val, "HOST", &h->Host);
			//printf("HOST: res = %d\n", res);
			if (res)
				break;
			/* TODO: parse HOST (here?) and if there is some crap in it, -404 */

			res = http_set_param(prm_name, prm_val, "USER-AGENT", &h->User_Agent);
			//printf("USER-AGENT: res = %d\n", res);
			if (res)
				break;

			res = http_set_param(prm_name, prm_val, "CONNECTION", &h->Connection);
			//printf("CONNECTION: res = %d\n", res);
			if (res)
				break;

			res = http_set_param(prm_name, prm_val, "ACCEPT", &h->Accept);
			//printf("ACCEPT: res = %d\n", res);
			if (res)
				break;

			res = http_set_param(prm_name, prm_val, "REFERER", &h->Referer);
			//printf("REFERER: res = %d\n", res);
			if (res)
				break;

			/* for POST method... */
			res = http_set_param(prm_name, prm_val, "CONTENT-LENGTH", &h->Content_Length);
			//printf("CONTENT-LENGTH: res = %d\n", res);
			if (res)
				break;

			res = http_set_param(prm_name, prm_val, "COOKIE", &h->Cookie);
			//printf("COOKIE: res = %d\n", res);
			if (res)
				break;

			/* This is for Content Generators */
			res = http_set_param(prm_name, prm_val, "X-DCGP-JOB-ID", &h->job_id);
			//printf("CONTENT-LENGTH: res = %d\n", res);
			if (res)
				break;

			/* Add more params right above this comment */

			res = http_set_param(prm_name, prm_val, "RANGE", &h->Range);
			// printf("RANGE: res = %d\n", res);
			break;

		} while (1);

		/* more than 1 parameter with same name */
		if (2 == res)
			return -400;

		s = l; /* go to next line */

	} while(1);

	if (!h->Host) // and HTTP 1.1!
		return -400;

//	// we have some body....
//	if (h->Content_Length)....??? what???
	return 0;
}


char *http_status[6][18] = {
	{
		"100 Continue",
		"101 Switching Protocols",
	},
	{
		"200 OK",
		"201 Created",
		"202 Accepted",
		"203 Non-Authoritative Information",
		"204 No Content",
		"205 Reset Content",
		"206 Partial Content",
	},
	{
		"300 Multiple Choices",
		"301 Moved Permanently",
		"302 Found",
		"303 See Other",
		"304 Not Modified",
		"305 Use Proxy",
		"306 (Unused)",
		"307 Temporary Redirect",
	},
	{
		"400 Bad Request",
		"401 Unauthorized",
		"402 Payment Required",
		"403 Forbidden",
		"404 Not Found",
		"405 Method Not Allowed",
		"406 Not Acceptable",
		"407 Proxy Authentication Required",
		"408 Request Timeout",
		"409 Conflict",
		"410 Gone",
		"411 Length Required",
		"412 Precondition Failed",
		"413 Request Entity Too Large",
		"414 Request-URI Too Long",
		"415 Unsupported Media Type",
		"416 Requested Range Not Satisfiable",
		"417 Expectation Failed",
	},
	{
		"500 Internal Server Error",
		"501 Not Implemented",
		"502 Bad Gateway",
		"503 Service Unavailable",
		"504 Gateway Timeout",
		"505 HTTP Version Not Supported",
	},

	/* DCGI extension */
	{
		"600 Content Generator Accepted",
		"601 Content Generator Rejected",
		"602 Too Many Content Generators Already",
		"603 Content Generator Name Too Long",
		"604 Bad Content Generator Name"
	}
};


/*
general-header

-Cache-Control
+ Connection
+ Date
-Pragma
-Trailer
-Transfer-Encoding
-Upgrade
-Via
-Warning

response-header

+ Accept-Ranges
? Age
? ETag
-Location
-Proxy-Authenticate
Retry-After (for 503)
+Server
?Vary
WWW-Authenticate (for 401)

entity-header
-Allow (must be in 405)
-Content-Encoding
-Content-Language
+Content-Length
-Content-Location
-Content-MD5
+Content-Range
+Content-Type
-Expires
+Last-Modified
*/

char *HTTP_GENERAL_HEADERS =
"HTTP/%d.%d %s\r\n\
Date: %s\r\n\
Connection: %s\r\n\
Server: swebs\r\n";

/* + Accept-Ranges: bytes\r\n"; */

char *HTTP_ENTITY_HEADERS =
"Content-Length: %d\r\n\
Content-Type: %s\r\n";

/*
  +Last-Modified: %s\r\n";
  +Content-Range: %d-%d/%d\r\n";
*/

char *con_close = "Close";
char *con_keep = "Keep-Alive";

int http_send_response(int sk, int code, time_t ts, char *ct, size_t cl, time_t lm, int *hl, size_t from, size_t to, size_t total, int close)
{
	char *response, *ptr;
	char *status = http_status[code/100 - 1][code % 100];
	char timestr[32], lm_timestr[32];
	int  len = 0, ret;
	char *cnx;

	rfc1123_date(ts, timestr);
	rfc1123_date(lm, lm_timestr);

	ptr = response = insistent_zmalloc(MAX_RESP_LEN);

	cnx = close?con_close:con_keep;

	sprintf(
		ptr,
		HTTP_GENERAL_HEADERS,
		1, 1,
		status,
		timestr,
		cnx
	);
	ptr += strlen(ptr);

	sprintf(ptr, "Accept-Ranges: bytes\r\n");
	ptr += strlen(ptr);

	sprintf(ptr, HTTP_ENTITY_HEADERS, cl, ct);
	ptr += strlen(ptr);

	sprintf(ptr, "Last-Modified: %s\r\n", lm_timestr);
	ptr += strlen(ptr);

	if (206 == code) {
		sprintf(ptr, "Content-Range: %d-%d/%d\r\n", from, to, total);
		ptr += strlen(ptr);
	}

	sprintf(ptr, "\r\n");

	len = strlen(response);
	if (hl)
		*hl = len;

	//ret = soket_write(sk, response, len);
	ret = send(sk, response, len, MSG_DONTWAIT | MSG_NOSIGNAL);

	free(response);

	return ret;
}


int http_send_builtin_response(int sk, int code, time_t ts, int *hl, int *bl, int close)
{
	char *response, *ptr;
	char *status = http_status[code/100 - 1][code % 100];
	char timestr[32];
	int len, ret;
	char *cnx;

	rfc1123_date(ts, timestr);
	ptr = response = insistent_zmalloc(MAX_RESP_LEN);

	cnx = close?con_close:con_keep;

	sprintf(
		ptr,
		HTTP_GENERAL_HEADERS,
		1, 1,
		status,
		timestr,
		cnx
	);
	ptr += strlen(ptr);

	sprintf(
		ptr,
		HTTP_ENTITY_HEADERS,
		strlen(status),
		"text/plain"
	);
	ptr += strlen(ptr);
	sprintf(ptr, "\r\n");

	len = strlen(response);
	if (hl)
		*hl = len;
	if (bl)
//		*bl = strlen(status) + 2;
		*bl = strlen(status);

	strcat(response, status);
	// strcat(response, "\r\n");

	len = strlen(response);
	// printf(" //// response len = %d\n", len);
	ret = send(sk, response, len, MSG_DONTWAIT | MSG_NOSIGNAL);
	free(response);
	return ret;
}

/* for DCGI output */
int http_read_response_header(int pipe, char *buf, int max, int timeo_sec)
{
	int bread;
	int len = 0;
	char *b = buf;
	char *hdrend = NULL;
	int ret = 0;
	int r;

	struct pollfd waitfor;
	int timeout = timeo_sec*1000; /* ms */
	time_t start;

	waitfor.fd = pipe;
	//waitfor.events = POLLIN | POLLERR | POLLHUP;
	waitfor.events = POLLIN;

	start = time(NULL);
	do {
		if (len == max)
			break;

		if ((time(NULL) - start) > (timeo_sec+1)) {
			ret = -504;
			break;
		};

		// revents = 0x0011 !!!!!! POLLHUP & POLLIN !!!!!
		r = poll(&waitfor, 1, timeout);
		if (0 == r) { /* timeout */
			ret = -504;
			break;
		};
		if (-1 == r) {
			perror("poll()");
			sleep(1);
			continue;
		};
		//bread = read(pipe, b + len, max - len);
		bread = read(pipe, b + len, 1); // :((
		if (-1 == bread) {
			perror("read(pipe)");
			continue;
		}
		if (0 == bread) {
			ret = 0;
			break;
		};
		len += bread;
		hdrend = strstr(b, "\r\n\r\n");
		/* we are not tolerant server, even to our own CGI :) */
		if (hdrend) {
			hdrend += strlen("\r\n\r\n");
			ret = hdrend - b;
			break;
		};
	} while (1);

	if (!hdrend && !ret)
		ret = -502;
	return ret;
}

int http_parse_response_header(char *data, struct http_response_header *h)
{
	return 0;
}
