
/*
 * HTTP protocol description:
 * http://www.ietf.org/rfc/rfc2616
 *
 * CGI description:
 * http://www.ietf.org/rfc/rfc3875
 */

#ifndef SWEBS_HTTP_H
#define SWEBS_HTTP_H

#define MAX_METH_LEN 8

#define MAX_FILE_LEN 256
#define MAX_PINF_LEN 256
#define MAX_ARGS_LEN 256
#define MAX_URL_LEN (MAX_FILE_LEN + MAX_PINF_LEN + MAX_ARGS_LEN)

#define MAX_HTTP_LEN 8

#define MAX_REQUEST_LINE_LEN (MAX_METH_LEN + MAX_URL_LEN + MAX_HTTP_LEN)

#define MAX_PARM_LEN 128
#define MAX_TOTAL_PARM_LEN 2048

#define MAX_HDR_LEN (MAX_METH_LEN + MAX_URL_LEN + MAX_TOTAL_PARM_LEN + 16)

typedef struct http_request_header {

	/* header buf */
	char buf[MAX_HDR_LEN + 1];
	char *offs;
	int len;

	/* request line */
	char request_line[MAX_REQUEST_LINE_LEN + 1];
	char *method;
	char *file;
	char *path_info;
	char *args;

	char *http;
	int ver;
	int subver;

	/* general */
	int code;
	int Cache_Control;

	char *Connection;

/*
	int Date;
	int Pragma;
	int Trailer;
	int Transfer_Encoding;
	int Upgrade;
	int Via;
	int Warning;
*/

	/* request */
	char *Accept;
/*
	int Accept_Charset;
	int Accept_Encoding;
	int Accept_Language;
	int Authorization;
	int Expect;
	int From;
*/

	char *Host;
/*
	int If_Match;
	int If_Modified_Since;
	int If_None_Match;
	int If_Range;
	int If_Unmodified_Since;
	int Max_Forwards;
	int Proxy_Authorization;
*/
	char *Range;
	char *Referer;
/*
	int TE;
*/
	char *User_Agent;

	/* entity */
	char *Content_Length;

/*
	int Allow;
	char *Content_Encoding;
	char *Content_Language;
	char *Content_Location;
	int Content_MD5;
	int Content_Range;
	int Content_Type;
	int Expires;
	int Last_Modified;
*/
	/* X- */
	char *xff;
	char *Cookie;

	char *job_id;

} HTTP_REQUEST_HEADER;

int http_recv_header(int sock, char *buf, int max, int timeo_sec);
int http_parse_request_header(char *data, struct http_request_header *h);
int http_bad_method(char *m);
void http_set_content_type(char *file, char *content_type);

#include <time.h>
int http_send_response(int sk, int code, time_t ts, char *ct, size_t cl, time_t lm, int *hl, size_t from, size_t to, size_t total, int ka);
int http_send_builtin_response(int sk, int code, time_t ts, int *hl, int *bl, int ka);

#define MAX_REASON_PHRASE_LEN 128
#define MAX_RESP_LEN 1024

typedef struct http_response_header {

	char buf[MAX_HDR_LEN+1];

	int have_status_line;
	int code;
	int ver;
	int subver;
	char reason_phrase[MAX_REASON_PHRASE_LEN];

	/* general */
	int Cache_Control;
	int Connection;
	int Date;
	int Pragma;
	int Trailer;
	int Transfer_Encoding;
	int Upgrade;
	int Via;
	int Warning;

	/* response */
	int Accept_Ranges;
	int Age;
	int ETag;
	char *Location;
	int Proxy_Authenticate;
	int Retry_After;
	int Server;
	int Vary;
	int WWW_Authenticate;

	/* entity */
	int Allow;
	int Content_Encoding;
	int Content_Language;
	int Content_Length;
	int Content_Location;
	int Content_MD5;
	int Content_Range;
	char *Content_Type;
	int Expires;
	int Last_Modified;
} HTTP_RESPONSE_HEADER;

int http_read_response_header(int pipe, char *buf, int max, int timeo_sec);
int http_parse_response_header(char *data, struct http_response_header *h);

#endif
