
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <sys/ioctl.h>
#include <string.h>
#include <sys/file.h>
#include <sys/epoll.h>
#include <sys/resource.h>

#include <pthread.h>

#include "utils.h"
#include "soket.h"
#include "http.h"
#include "swebs.h"
#include "fd-cache.h"

struct swebs_config config = {

	.tcp_port = TCP_PORT,
	.send_timeo = SEND_TIMEO_DFLT,
	.recv_timeo = RECV_TIMEO_DFLT,
	.fragment_size_KiB = 32,

/*
	.cgi_buf_size_Kb = 32,
	.cgi_data_timeo = CGI_DATA_TIMEOUT,
	.cgi_max_time = CGI_MAX_TIME
*/
};

/* 0-th element is unused! */
struct cong congs[MAX_CONGS];
unsigned long long int addr_mask = 0;
struct fd_cache *fdc;

void* acceptor_th(void* arg)
{
	struct worker *me = (struct worker*)arg;
	int lsk = me->in->rdfd;
	struct job_queue *to_fetcher = me->out;
	struct job_queue *to_congmux = me->aux;

	struct job *job;
	struct connection *cnx;

	int r;

	int epfd;
	struct epoll_event poll_for, event;
	int bytes;
	int cid;

	epfd = epoll_create(512);
	if (-1 == epfd) {
		perror("epoll_create()");
		exit(1);
	}

	poll_for.events = EPOLLIN;
	poll_for.data.ptr = NULL;
	r = epoll_ctl(epfd, EPOLL_CTL_ADD, lsk, &poll_for);
	if (-1 == r) {
		perror("epoll_ctl()");
		exit(1);
	}

	/* printf("%s started.\n", me->name); */

	for (;;) {

		event.events = 0;
		r = epoll_wait(epfd, &event, 1, 1000);
		if (!r) {
			// здесь можно проверять таймаут на поступление запроса.
			// тогда надо засекать время подключения.
			continue;
		}
		if (r < 0) {
			perror("epoll_wait() in ACCEPTOR");
			// continue;
			exit(1);
		}

		cnx = (struct connection*)event.data.ptr;
		//printf("epoll: r = %d, cnx = %p, sk = %d, ev 0x%x\n", r, cnx, cnx?cnx->sock:lsk, event.events);

		/* new client (or cong) */
		if (!cnx) {
			cnx = swebs_new_connection(lsk, &config, epfd);
			// printf("CLIENT from %s, cnx @ %p\n", cnx->who, cnx);
			continue;
		};

		if ((event.events & EPOLLHUP) || (event.events & EPOLLERR)) {

//			printf("#1 ... EPOLLIN, but no data on %d, closing!\n", cnx->sock);
			/* safe? */
			epoll_ctl(epfd, EPOLL_CTL_DEL, cnx->sock, NULL);
			close(cnx->sock);

			// it seems we have memory leakage here... (???)
//			if (cnx->job) {
//				printf("CK PNT (freeing job) #1...\n");
//				free(job); // ???
				// поимели double free....
//			}

			// printf("cnx @ %p is to freed...\n", cnx);
			free(cnx);
			// а если все таки файл уже открыт?....
			continue;
		};

		/* check if we _really_ have data */
		r = ioctl(cnx->sock, FIONREAD, &bytes);
		if (-1 == r) {
			printf("ioctl(FIONREAD) failed\n");
			perror("ioctl(FIONREAD)");
			exit(1);
		}

		if (!bytes) {
//			printf("#2... EPOLLIN, but no data on %d, closing!\n", cnx->sock);
			/* safe? */
			epoll_ctl(epfd, EPOLL_CTL_DEL, cnx->sock, NULL);
			close(cnx->sock);

			// it seems we have memory leakage here... (???)
//			if (cnx->job) {
//				printf("CK PNT (freeing job) #2...\n");
//				free(job); ///???
				// поимели double free... при закрытии браузера...
//			}
			// printf("cnx @ %p is to freed...\n", cnx);
			free(cnx);

			// а если все таки файл уже открыт?....
			continue;
		}

		/*
		 * we did not receive headers yet,
		 * don't create new job, continue polling
		 */
		if ((cnx->job) && (!cnx->have_request)) {
			job = cnx->job;
			goto __read_more;
		}

		job = cnx->job = swebs_new_job();
		job->cnx = cnx;
		job->sock = cnx->sock;
		strcpy(job->who, cnx->who);
		job->pollinfd = epfd;

		job->num = ++cnx->job_num;
		cnx->have_request = 0;
		//printf("REQUEST from %s for CNX @ %p, job @ %p\n", cnx->who, cnx, job);

              __read_more:
		//r = swebs_get_request(job, &config);
		r = swebs_get_request(job);
		// printf("swebs_get_request(): r = %d\n", r);

		/* no '\r\n\r\n yet '*/
		if (!r)
			continue;

		if (-1 == r) { // should never happen (?)
			printf("OOPS!!! read() from %d failed!\n", job->sock);
			/* ? */
			epoll_ctl(epfd, EPOLL_CTL_DEL, cnx->sock, NULL);
			job->done = 1;
			job->is_last = 1;
			goto __done;
			/* ? */
		}

		cnx->have_request = 1;

		if (r < 0) {
			job->status = -r;
			goto __http_error;
		}

		swebs_parse_http_request(job);
//		printf("%s from %s wants '%s' on sock %d (rq #%d)\n", job->hdr.User_Agent, job->who, job->hdr.request_line, job->sock, job->num);
//		printf("%s from %s wants '%s' on sock %d (rq #%d)\n", job->hdr.User_Agent, job->who, job->hdr.file, job->sock, job->num);

		/* not found "locally", try find some cong for that URL */
		if (job->status == 404) {
			/* try to find some cong for the request */
			printf("looking in '%s'... \n", job->hdr.file);
			cid = swebs_find_cong(congs, job->hdr.file);
			if (cid) { // request to a CONG
				// ???
				job->cong = &congs[cid];
				// congs[cid].cnx->job = job;
				epoll_ctl(epfd, EPOLL_CTL_DEL, job->sock, NULL);
				swebs_gen_jobid(job, addr_mask);
				job->status = 0;
				swebs_pass_job(to_congmux, job);
				continue;
			}
		}

              __http_error:
		if ((job->status) && (job->status < 600)) {
			swebs_send_builtin_response(job);
			// check error?
			job->done = 1;
			goto __done;
		};

		/* DCGI request */
		if (job->status == 600) { // new ConG

			printf("NEW CONG!!!\n\n");
			epoll_ctl(epfd, EPOLL_CTL_DEL, job->sock, NULL);
			cid = swebs_ins_cong(congs, job->hdr.file, cnx);

			if (!cid) {
				job->status = 603; // or 601
				job->done = 1;
				job->is_last = 1; // ????
			}

			job->cong = &congs[cid];
			/* notify the cong */
			swebs_send_builtin_response(job);
			/* let congmux know about new generator */
			swebs_pass_job(to_congmux, job);
			continue;
		}

		/* убрать!!! */
//		if (job->finfo.st_mode & S_IXUSR) {
//			job->status = 404;
//			goto __http_error;
//		};

		if (job->hdr.Range)
			job->status = 206;
		if (!job->status)
			job->status = 200;

		r = swebs_send_response(job);
		if (-1 == r) {
			epoll_ctl(epfd, EPOLL_CTL_DEL, cnx->sock, NULL);
			job->done = 1;
			job->is_last = 1;
			goto __done;
		}

		if ((!job->response_blen) || (job->hdr.method[0] == 'H')) {
			job->done = 1;
			goto __done;
		}

		// это надо делать до того как слать ответ... чтобы послать 503 например
		job->fd = swebs_prepare_to_load_file(job, fdc, &config);
		if (-1 == job->fd) {
			job->done = 1;
			job->is_last = 1;
		}

	      __done:

		/* stop polling the sock until the job is logged (or file sent?...) */
		r = epoll_ctl(epfd, EPOLL_CTL_DEL, job->sock, NULL);
		CHECK_AND_EXIT(r != -1);
		r = swebs_pass_job(to_fetcher, job);
		CHECK_AND_EXIT(r == sizeof(struct job*));
	};
}

void* congmux_th(void *arg)
{
	struct worker *me = (struct worker*)arg;
	struct job_queue *from_parser = me->in;
	// struct job_queue *to_sender = me->out;
	struct job_queue *to_logger = me->aux;

	struct job *job;
	struct cong *cong;
	int epfd;
	struct epoll_event poll4;
	int r;
	int cong_sock;

//	struct connection *cnx;

	epfd = epoll_create(512);
	if (-1 == epfd) {
		perror("epoll_create()");
		exit(1);
	}

	poll4.events = EPOLLIN;
	poll4.data.ptr = NULL;
	r = epoll_ctl(epfd, EPOLL_CTL_ADD, from_parser->rdfd, &poll4);
	if (-1 == r) {
		perror("epoll_ctl()");
		exit(1);
	}

	// printf("%s started.\n", me->name);

	for (;;) {

		poll4.events = 0;
		/* poll with 1 sec timeout... just in case */
		r = epoll_wait(epfd, &poll4, 1, 1000);
		if (!r) /* timeout */
			continue;
		if (r < 0) {
			perror("epoll_wait() in CONGMUX");
			//continue;
			exit(1);
		}

		job  = (struct job*)poll4.data.ptr;
		cong = (struct cong*)job;

		printf("CONG_MUX has something: events 0x%x, data %p\n", poll4.events, poll4.data.ptr);

		if (!job) { /* job from parser */

			job = swebs_get_job(from_parser);
			CHECK_AND_EXIT(job);

			if (job->done);;;;;;;
			if (job->is_last);;;;;;;;

			printf("CONG MUXER: job from PARSER\n");

			/* new Cong connected */
			if (job->status == 600) {
				/* cong sock is in job->sock */
				printf("CONG MUXER: new CONG (%s) from %s\n", job->hdr.file, job->who);
				poll4.events = EPOLLIN;
				poll4.data.ptr = job->cong;
				r = epoll_ctl(epfd, EPOLL_CTL_ADD, job->sock, &poll4);
				if (-1 == r) {
					perror("epoll_ctl()");
					exit(1);
				}
				r = swebs_pass_job(to_logger, job);
				CHECK_AND_EXIT(r == sizeof(struct job*));
				continue;
			}

			/* request to a Cong... */
			printf("CONG MUXER: some job for CONG #%d (%s) @%p\n", job->cong->n, job->cong->name, job->cong);

			cong_sock = job->cong->cnx->sock;

			// всю эту пургу <<<< надо куда нить заопределить.

			// prepare buf
			job->buf = malloc(8192);
			memset(job->buf, 0, 8192);
			// dont forget to free!!!!!!!! самуэре, балиать

			strcat((char*)job->buf, job->hdr.request_line);
			strcat((char*)job->buf, "\r\n");

			if (job->hdr.Referer) {
				strcat((char*)job->buf, "Referer: ");
				strcat((char*)job->buf, job->hdr.Referer);
				strcat((char*)job->buf, "\r\n");
			};

			if (job->hdr.User_Agent) {
				strcat((char*)job->buf, "User-Agent: ");
				strcat((char*)job->buf, job->hdr.User_Agent);
				strcat((char*)job->buf, "\r\n");
			};

			if (job->hdr.Range) {
				strcat((char*)job->buf, "Range: ");
				strcat((char*)job->buf, job->hdr.Range);
				strcat((char*)job->buf, "\r\n");
			};

			if (job->hdr.Content_Length) {
				strcat((char*)job->buf, "Content-Length: ");
				strcat((char*)job->buf, job->hdr.Content_Length);
				strcat((char*)job->buf, "\r\n");
			};

			strcat((char*)job->buf, "X-DCGP-JOB-ID: ");
			strcat((char*)job->buf, job->id);
			strcat((char*)job->buf, "\r\n\r\n");

			r = send(cong_sock, job->buf, strlen((char*)job->buf), MSG_NOSIGNAL | MSG_DONTWAIT);
			if (-1 == r);;;;
			if (r != strlen((char*)job->buf));;;; // what???

			// >>>>>> конец пурге

			continue;
		}

		// handle EPOLLERR & EPOLLHUP here!

		/* data from cong (?) однозначно? */
		if (poll4.events & EPOLLIN) {

			char frag_hdr[FRAG_HDR_LEN + 1];
			char *jobid, *len;
			size_t bytes = 0;
			unsigned long long flen;

			printf("CONG MUXER: data from CONG\n");

			cong_sock = cong->cnx->sock;

			r = ioctl(cong_sock, FIONREAD, &bytes);
			CHECK_AND_EXIT(r != -1);

			if (!bytes) {
				printf("CONG gone...\n");
				printf("Closing %d\n", cong_sock);
				/* safe? */
				epoll_ctl(epfd, EPOLL_CTL_DEL, cong_sock, NULL);
				close(cong_sock);
				printf("cnx @ %p freed\n", cong->cnx);
				free(cong->cnx);
				swebs_del_cong(congs, cong->name); //....
				// а чо если он отвалилсо пока клиента обслуживал?...
				// как это определить и чо делать c job и с клиентом?
				continue;
				// exit(1);
			}

			printf("%d bytes\n", bytes);

//			if (bytes < FRAG_HDR_LEN) { // bad cong?... кокнуть его?.... если продолжать, будет непонятно, что это и для кого....
//				printf("have only %d bytes... \n", bytes);
//				continue;
//			}

			memset(frag_hdr, 0, sizeof(frag_hdr));

			/* get jobid:len and parse */
			r = recv(cong_sock, frag_hdr, FRAG_HDR_LEN, MSG_WAITALL | MSG_NOSIGNAL);
			// читаем, смотрим, сколько указано байт, сравниваем с (bytes - FRAG_HDR_LEN)
			printf("frag hdr: '%s'\n", frag_hdr);

			if (-1 == r);;;;;;;;
			if (r < FRAG_HDR_LEN);;;;;

			if (frag_hdr[63] != '\n');;;;;
			frag_hdr[63] = 0;

			if (frag_hdr[62] != '\r');;;;;
			frag_hdr[62] = 0;

			if (frag_hdr[45] != ':');;;;
			frag_hdr[45] = 0;

			len = &frag_hdr[46];
			jobid = frag_hdr;

			if (jobid[6] != '$');;;;
			if (jobid[29] != '$');;;;

			printf("id: '%s', len: '%s'\n", jobid, len);

			r = sscanf(len, "%llx", &flen);
			if (r != 1);;;;;
			printf("len = %lld (scanned %s)\n", flen, len);

			// сравнить flen и из ioctl(FIONREAD)...

			job = swebs_check_jobid(jobid, addr_mask);
			if (!job) {
				printf("BAD PTR!\n");
				exit(1);
			};;;;;

			job->buf_total = flen;

			// job is done!
			if (!job->buf_total);;;;;;

			memset(job->buf, 0, 8192);
			r = recv(cong_sock, job->buf, flen, MSG_NOSIGNAL | MSG_WAITALL);
			if (-1 == r);;;;;;;;
			if (r < job->buf_total);;;;;;;;
			//printf("frag from cong: '%s'", job->buf);

			// start poll client for readiness to accept data
			poll4.events = EPOLLOUT;
			poll4.data.ptr = job;
			r = epoll_ctl(epfd, EPOLL_CTL_ADD, job->sock, &poll4);
			if (-1 == r) {
				perror("epoll_ctl(ADD)");
				exit(1);
			}

			// а потом кода иво вертать взад?
			r = epoll_ctl(epfd, EPOLL_CTL_DEL, cong_sock, NULL);
			if (-1 == r) {
				perror("epoll_ctl(DEL)");
				exit(1);
			}

			continue;
		}

		if (poll4.events & EPOLLOUT) {

			int w;

			printf("sending '%s' to %d...\n", job->buf, job->sock);
			w = send(job->sock, job->buf + job->buf_sent, job->buf_total - job->buf_sent, MSG_NOSIGNAL | MSG_DONTWAIT);
			// handle partial writes, like in sender!

			r = epoll_ctl(epfd, EPOLL_CTL_DEL, job->sock, NULL);
			if (-1 == r) {
				perror("epoll_ctl(DEL)");
				exit(1);
			}

			///////////////////
			// а потом кода иво вертать взад?
			// start poll client for readiness to accept data
			poll4.events = EPOLLIN;
			poll4.data.ptr = job->cong;
			r = epoll_ctl(epfd, EPOLL_CTL_ADD, job->cong->cnx->sock, &poll4);
			if (-1 == r) {
				perror("epoll_ctl(ADD)");
				exit(1);
			}
			//////////////////////

			free(job->buf); // not here!!!!!!!!

			r = swebs_pass_job(to_logger, job);
			CHECK_AND_EXIT(r == sizeof(struct job*));
			// free(job);
		}
	}
}

void* fetcher_th(void *arg)
{
	struct worker *me = (struct worker*)arg;
	struct job_queue *from_parser_or_sender = me->in;
	struct job_queue *to_sender = me->out;

	struct job *job;
	int r;

//	printf("%s started.\n", me->name);

	for (;;) {

		job = swebs_get_job(from_parser_or_sender);
		CHECK_AND_EXIT(job);

		if (job->done)
			goto __done;

//		printf("F.\n"); 

		r = swebs_get_fragment(job, &config);
		if (r > 0)
			job->buf_sent = 0;
		else
			job->done = 1;

	      __done:
		r = swebs_pass_job(to_sender, job);
		CHECK_AND_EXIT(r == sizeof(struct job*));
	};
}

void* sender_th(void *arg)
{
	struct worker *me = (struct worker*)arg;
	struct job_queue *from_fetcher = me->in;
	struct job_queue *to_fetcher = me->aux;
	struct job_queue *to_logger = me->out;

	struct job *job;
	int done, r;
	int epfd;
	struct epoll_event poll4;

	epfd = epoll_create(512);
	if (-1 == epfd) {
		perror("epoll_create()");
		exit(1);
	};

	poll4.events = EPOLLIN;
	poll4.data.ptr = NULL;
	r = epoll_ctl(epfd, EPOLL_CTL_ADD, from_fetcher->rdfd, &poll4);
	if (-1 == r) {
		perror("epoll_ctl()");
		exit(1);
	}

//	printf("%s started.\n", me->name);

	for (;;) {

		poll4.events = 0;
		/* poll with 1 sec timeout... just in case */
		r = epoll_wait(epfd, &poll4, 1, 1000);
		if (!r) /* timeout */
			continue;
		if (r < 0) {
			perror("epoll_wait() in SENDER");
			//continue;
			exit(1);
		}

		job = poll4.data.ptr;

		if (!job) { /* message from fetcher, fragment is loaded */

			job = swebs_get_job(from_fetcher);
			CHECK_AND_EXIT(job);

			if (job->done)
				goto __done;

			poll4.events = EPOLLOUT;
			poll4.data.ptr = job;
			r = epoll_ctl(epfd, EPOLL_CTL_ADD, job->sock, &poll4);
			// what if job->sock already CLOSED?
			if (-1 == r) {
				printf("JOB @ %p, epoll_ctl(epfd %d, EPOLL_CTL_ADD, sock %d, ...) failed\n", job, epfd, job->sock);
				perror("epoll_ctl(EPOLL_CTL_ADD, new sock for EPOLLOUT)");
				exit(1);
			}
			continue;
		}

		/* socket ready for writing */
		done = swebs_send(job);

		if (!done) {

			/* partial write, wait for EPOLLOUT again */
			if (job->buf_sent != job->buf_total) {
				// printf("---$$$ %d sent\n", job->buf_sent);
				continue;
			}

			/* Buffer sent,
			 * stop polling this socket for a while,
			 * otherwise EPOLLOUT may be received before next fragment is loaded
			 */

			r = epoll_ctl(epfd, EPOLL_CTL_DEL, job->sock, NULL);
			if (-1 == r) {
				printf("-- epoll_ctl(EPOLL_CTL_DEL)\n");
				perror("-- epoll_ctl(EPOLL_CTL_DEL)");
				exit(1);
			}

			r = swebs_pass_job(to_fetcher, job);
			CHECK_AND_EXIT(r == sizeof(struct job*));
			continue;
		}

	      __done:
		r = epoll_ctl(epfd, EPOLL_CTL_DEL, job->sock, NULL);
		if (job->is_last) {
			close(job->sock);
			// printf("cnx @ %p freed\n", job->cnx);
			free(job->cnx);
		};
		swebs_end_send_file(job, fdc);
		r = swebs_pass_job(to_logger, job);
		CHECK_AND_EXIT(r == sizeof(struct job*));
	}
}

void* logger_th(void* arg)
{
	struct worker *me = (struct worker*)arg;
	struct job_queue *from_sender = me->in;
	int log = me->out->wrfd;
	struct job *job;

	int r;
	struct epoll_event poll4;

//	printf("%s started.\n", me->name); 

	for (;;) {

		job = swebs_get_job(from_sender);
		CHECK_AND_EXIT(job);
/*
		printf(
			"logger : '%s'[%d] (%d+%d/%d) LOGGED [%d]\n",
			job->hdr.request_line,
			job->num,
			job->response_hlen,
			job->sent,
			job->response_blen,
			job->status
		);
*/
		swebs_record_log(log, job);

		if (job->is_last)
			goto __free_job;

		/* new cong, don't add into Acceptor epoll set */
		if (job->status >= 600)
			goto __free_job;

		/* more requests in the connection are expected
		   Q: move this to sender? */
		poll4.events = EPOLLIN;
		poll4.data.ptr = job->cnx;
		r = epoll_ctl(job->pollinfd, EPOLL_CTL_ADD, job->sock, &poll4);
		if (-1 == r) {
			perror("epoll_ctl(EPOLL_CTL_ADD)\n");
			exit(1);
		}

	      __free_job:
		// printf("job @ %p freed\n", job);
		free(job);
	};
}

struct job_queue links[] = {

	{ /* this one is special */
		.to_whom = "The-World",
		.rdfd = -1,
		.wrfd = -1,
		.load = 0
	},

	{
		.to_whom = "To-CongMux",
		.rdfd = 0,
		.wrfd = 0,
		.load = 0
	},
	{
		.to_whom = "To-Fetcher",
		.rdfd = 0,
		.wrfd = 0,
		.load = 0
	},
	{
		.to_whom = "To-Sender",
		.rdfd = 0,
		.wrfd = 0,
		.load = 0
	},
	{
		.to_whom = "To-Logger",
		.rdfd = 0,
		.wrfd = 0,
		.load = 0
	},

	{ /* this one is special too */
		.to_whom = "Memory",
		.rdfd = -1,
		.wrfd = -1,
		.load = 0
	},
};

#define NLINKS \
	(sizeof(links)/sizeof(struct job_queue))

struct worker workers[] = {
	{
		.name	= "Acceptor",
		.action = acceptor_th,
		.in_queue_name = "The-World",
		.out_queue_name = "To-Fetcher",
		.aux_queue_name = "To-CongMux"
	},
	{
		.name	= "CongMux",
		.action = congmux_th,
		.in_queue_name = "To-CongMux",
		.out_queue_name = "To-Sender",
		.aux_queue_name = "To-Logger"
	},
	{
		.name	= "Fetcher",
		.action = fetcher_th,
		.in_queue_name = "To-Fetcher",
		.out_queue_name = "To-Sender",
		.aux_queue_name = NULL
	},
	{
		.name	= "Sender",
		.action = sender_th,
		.in_queue_name = "To-Sender",
		.out_queue_name = "To-Logger",
		.aux_queue_name = "To-Fetcher"
	},
	{
		.name	= "Logger",
		.action = logger_th,
		.in_queue_name = "To-Logger",
		.out_queue_name = "Memory",
		.aux_queue_name = NULL
	}
};

#define NWORKERS \
	sizeof(workers)/sizeof(struct worker)

int lsock;

/* docroot is relative to HOME envar */
char *docroot = "/www";

/* these two are relative to docroot */
char *logfile = ".log/access.log";
char *cgierrfile = ".log/cgierr.log";
char *syserrfile = ".log/syserr.log";
char *confile = ".conf/config";

int acclog;

// int cgierrlog;
// int syserrlog;

int main(int argc, char *argv[], char *env[])
{
	int wn;
	struct rlimit nofile_limit;
//	int r;

	getrlimit(RLIMIT_NOFILE, &nofile_limit);
	printf("max file soft limit = %d, max file hard limit = %d\n", (int)nofile_limit.rlim_cur, (int)nofile_limit.rlim_max);
	// nofile_limit.rlim_cur = 8192;
	// nofile_limit.rlim_max = 8192;
	// r = setrlimit(RLIMIT_NOFILE, &nofile_limit);
	// if (-1 == r) {
	//	perror("setrlimit(RLIMIT_NOFILE)");
	//}

	// printf("RAND_MAX = %d\n", RAND_MAX);
	swebs_generate_addr_mask(&addr_mask);

	memset(congs, 0, sizeof(congs));

	swebs_check_uids();
	swebs_goto_wrkdir(docroot);

	/* turn the memory on */
	acclog = swebs_open_logfile(logfile);
	for (wn = 0; wn < NWORKERS; wn++) {
		if (!strcmp(workers[wn].name, "Logger"))
			break;
	}
	workers[wn].out = &links[NLINKS-1];
	workers[wn].out->wrfd = acclog;

/*
	cgierrlog = swebs_open_logfile(cgierrfile);
	stages[LAUNCHER].pipes.log_out = cgierrlog;
*/

	swebs_load_config(confile, &config);
	fdc = fdc_init(0);

	/* connect to the world */
	lsock = swebs_listen_on(config.tcp_port);
	for (wn = 0; wn < NWORKERS; wn++) {
		if (!strcmp(workers[wn].name, "Acceptor"))
			break;
	}
	workers[wn].in = &links[0];
	workers[wn].in->rdfd = lsock;

	/* create queues (pipes) between stages and link the stages */
	swebs_create_staff(workers, NWORKERS, links, NLINKS);

//	printf("sizeof(JOB) = %d\n", sizeof(struct job));

/*
	syserrlog = swebs_open_logfile(syserrfile);
	dup2(syserrlog, STDERR_FILENO); // ?
*/

	swebs_show_stat(links, NLINKS, fdc);
	return 0;
}
