
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "soket.h"

/*
 * soket_server() creates a IPV4/TCP socket,
 * sets REUSEADDR for it,
 * binds it to __any__ interface and specified port,
 * and, finally, make it listening
 *
 * returns sk created;
 */

int soket_server(unsigned short port, int accept_queue_len)
{
	int sk, err;
	struct sockaddr_in our_addr;
	int on = 1;

	sk = socket(AF_INET, SOCK_STREAM, 0);
	if ( -1 == sk )
		return sk;

	err = setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int));
	if ( -1 == err )
		return err;

	memset(&our_addr, 0, sizeof(our_addr));
	our_addr.sin_family = AF_INET;
	our_addr.sin_addr.s_addr = INADDR_ANY;
	our_addr.sin_port = htons(port);

	err = bind(sk, (struct sockaddr*)&our_addr, sizeof(our_addr));
	if ( -1 == err )
		return err;

	err = listen(sk, accept_queue_len);
	if ( -1 == err )
		return err;

	return sk;
}

int soket_accept(int sk, char *ip, struct timeval *sndtimeo)
{
	int dsk;
	struct sockaddr_in addr;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	int r, k;

	for (k = 3; k; k--) {
		// printf("BEFORE ACCEPT: skfd = %d, addrlen = %d\n", sk, addrlen);
		dsk = accept(sk, (struct sockaddr*)&addr, &addrlen);
		// printf("dsk = %d\n", dsk);
		if (dsk > 0)
			break;
		if (!dsk) {
			printf("BUG!!! (fd #0 was errorneously closed)\n");
			exit(1);
		}
		perror("accept()");
		sleep(1);
	};

	if (-1 == dsk)
		return dsk;

	// printf("AFTER ACCEPT: skfd = %d, dsk = %d\n", sk, dsk);
	if (sndtimeo) {
		r = setsockopt(dsk, SOL_SOCKET, SO_SNDTIMEO, sndtimeo, sizeof(struct timeval));
		if (-1 == r)
			perror("setsockopt(SO_SNDTIMEO)");
	};

	inet_ntop(AF_INET, &addr.sin_addr, ip, IP_STR_LEN);
	// sin_port....
	return dsk;
}

int soket_client(char *srv)
{
	int sk, err = -1;
	char *host_str, *port_str;
	int port;
	int r;
	struct sockaddr_in serv_addr;
	struct in_addr host_ip;
	struct hostent *hostent;
	char service[64];

//	struct sockaddr_in our_addr;
//	int on = 1;

	strcpy(service, srv);

	printf("Creating TCP socket...");
	fflush(stdout);
	sk = socket(AF_INET, SOCK_STREAM, 0);
	if (-1 == sk)
		return sk;
	printf("OK, sk %d\n", sk);

	port_str = index(service, ':');
	if (!port_str) {
		// printf("NO ':'\n");
		return err;
	}
	// printf(": at %p\n", port_str);

	port_str[0] = 0; /* split host:port */
	port_str++;

	r = sscanf(port_str, "%d", &port);
	if (r != 1)
		return err;

	host_str = service;
	r = inet_aton(host_str, &host_ip);

	if (!r) {
		printf("Resolving '%s'...", host_str);
		fflush(stdout);
		hostent = gethostbyname(host_str);
		if (!hostent) {
			herror("gethostbyname()");
			return -1;
		}
		//printf("OK, %.4x\n", *((struct in_addr *)hostent->h_addr));
		// host_ip = hostent->h_addr_list[0];
		host_ip = *((struct in_addr *)hostent->h_addr);
	}

	memset(&serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);
	serv_addr.sin_addr = host_ip;

	printf("Connecting to 0x%.4x... \n", host_ip.s_addr);
	fflush(stdout);
	err = connect(sk, (struct sockaddr*)&serv_addr, sizeof(struct sockaddr_in));
	if (-1 == err) {
		perror("connect()");
		return err;
	}
	printf("OK\n");

	return sk;
}

/*
int soket_read(int sk, unsigned char *buf, unsigned long len)
{
	return read(sk, buf, len);
}

int soket_close(int sk)
{
	int err = close(sk);
	if (-1 == err)
		perror("close()");
	return err;
}
*/
