
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>
#include <sys/file.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <curl/curl.h>

#include "soket.h"
#include "utils.h"
#include "http.h"
#include "swebs.h"
#include "fd-cache.h"

void swebs_check_uids(void)
{
	uid_t uid  = getuid();
	uid_t euid = geteuid();

	if ( !uid || !euid ) {
		fprintf(stderr, "We don't want to run with root privilegies (uid = %d, euid = %d)\n", uid, euid);
		exit(1);
	};
}

void swebs_goto_wrkdir(char *reldocroot)
{
	char docroot[128];
	char *home;
	int err;

	home = getenv("HOME");
	if (!home) {
		printf("no homedir...\n");
		exit(1);
	};
	memset(docroot, 0, sizeof(docroot));
	strcpy(docroot, home);
	strcat(docroot, "/www");
	err = chdir(docroot);
	if (-1 == err) {
		perror("chdir");
		exit(1);
	};
}

int swebs_open_logfile(char *logname)
{
	int fd;

	printf("Opening log (%s)... ", logname);
	fflush(stdout);
	fd = open(logname, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR);
	if (-1 == fd) {
		perror("open()");
		exit(1);
	};
	printf("OK\n");
	return fd;
}

void swebs_load_config(char *fname, struct swebs_config *cfg)
{
	FILE *s;
	int line_count = 0;
	#define MAX_CFG_LEN 128
	char buf[MAX_CFG_LEN];
	char buf1[MAX_CFG_LEN];
	char buf2[MAX_CFG_LEN];
	int num;

	printf("Parsing config file (%s)... \n", fname);
	s = fopen(fname, "r");
	if (!s) {
		// fprintf(stderr, "proceeding with built-in config..."); чонить в этом роде
		perror("fopen");
		return;
	};

	while (!feof(s)) {

		memset(buf,  0, MAX_CFG_LEN);
		memset(buf1, 0, MAX_CFG_LEN);
		memset(buf2, 0, MAX_CFG_LEN);

		fgets(buf, MAX_CFG_LEN, s);
		line_count++;

		/* comment */
		if (sscanf(buf, " # %s ", buf1) == 1)
			continue;

		/* empty or incorrect line */
		if (sscanf(buf, " %s = %s ", buf1, buf2) != 2) {
			// printf("error in line %d // %s\n", line_count, buf);
			continue;
		};

		// printf("line %d: %s", line_count, buf);
		if (sscanf(buf, " tcp_port = %d ", &num) == 1) {
			cfg->tcp_port = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else if (sscanf(buf, " send_timeo = %d ", &num) == 1) {
			cfg->send_timeo = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else if (sscanf(buf, " recv_timeo = %d ", &num) == 1) {
			cfg->recv_timeo = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else if (sscanf(buf, " fragment_size = %d ", &num) == 1) {
			cfg->fragment_size_KiB = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else if (sscanf(buf, " cgi_buf_size = %d ", &num) == 1) {
			cfg->cgi_buf_size_KiB = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else if (sscanf(buf, " cgi_data_timeo = %d ", &num) == 1) {
			if ((num < 1) || (num > 60)) {
				printf("WARNING: bad 'cgi_data_timeo', using %d\n",  5);
				num = 5;
			};
			cfg->cgi_data_timeo = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else if (sscanf(buf, " cgi_max_time = %d ", &num) == 1) {
			if ((num < 1) || (num > 120)) {
				printf("WARNING: bad 'cgi_max_time', using %d\n",  60);
				num = 60;
			};
			cfg->cgi_max_time = num;
			printf("'%s' = %s\n", buf1, buf2);
		} else
			printf("line %d ignored :) // %s", line_count, buf);
	};
	fclose(s);
}

void swebs_create_staff(struct worker *workers, int nworkers, struct job_queue *links, int nlinks)
{
	int wn, ln;
	int err;
	pthread_t *staff;
	int p[2];

	for (ln = 0; ln < nlinks; ln++) {

		if ( (links[ln].rdfd == -1) || (links[ln].wrfd == -1))
			continue;

		printf("Creating '%s' queue...", links[ln].to_whom);
		fflush(stdout);

		err = pipe(p);
		if (-1 == err) {
			perror("pipe()");
			exit(1);
		};

		links[ln].rdfd = p[PIPE_RD_END];
		links[ln].wrfd = p[PIPE_WR_END];
		printf("OK\n");

		for (wn = 0; wn < nworkers; wn++) {

			if (workers[wn].in_queue_name)
				if (!strcmp(workers[wn].in_queue_name, links[ln].to_whom))
					workers[wn].in = &links[ln];

			if (workers[wn].out_queue_name)
				if (!strcmp(workers[wn].out_queue_name, links[ln].to_whom))
					workers[wn].out = &links[ln];

			if (workers[wn].aux_queue_name)
				if (!strcmp(workers[wn].aux_queue_name, links[ln].to_whom))
					workers[wn].aux = &links[ln];
		}
	}

	printf("Allocating working places... ");
	fflush(stdout);
	staff = (pthread_t*)malloc(nworkers*sizeof(pthread_t));
	if (!staff) {
		perror("failed");
		exit(1);
	}
	printf("OK\n");

	for (wn = 0; wn < nworkers; wn++) {
		printf("Hiring Mr. %s... ", workers[wn].name);
		fflush(stdout);
		err = pthread_create(&staff[wn], NULL, workers[wn].action, &workers[wn]);
		if (err) {
			perror("pthread_create()");
			exit(1);
		};
		printf("OK\n");
	}
}

/*
4.1. Request Meta-Variables . . . . . . . . . . . . . . . . .  10

-4.1.1.  AUTH_TYPE. . . . . . . . . . . . . . . . . . . .  11
-4.1.2.  CONTENT_LENGTH . . . . . . . . . . . . . . . . .  12
4.1.3.  CONTENT_TYPE . . . . . . . . . . . . . . . . . .  12
+4.1.4.  GATEWAY_INTERFACE. . . . . . . . . . . . . . . .  13
+4.1.5.  PATH_INFO. . . . . . . . . . . . . . . . . . . .  13
-4.1.6.  PATH_TRANSLATED. . . . . . . . . . . . . . . . .  14
+4.1.7.  QUERY_STRING . . . . . . . . . . . . . . . . . .  15
+4.1.8.  REMOTE_ADDR. . . . . . . . . . . . . . . . . . .  15
-4.1.9.  REMOTE_HOST. . . . . . . . . . . . . . . . . . .  16
-4.1.10. REMOTE_IDENT . . . . . . . . . . . . . . . . . .  16
-4.1.11. REMOTE_USER. . . . . . . . . . . . . . . . . . .  16
+4.1.12. REQUEST_METHOD . . . . . . . . . . . . . . . . .  17
+4.1.13. SCRIPT_NAME. . . . . . . . . . . . . . . . . . .  17
+4.1.14. SERVER_NAME. . . . . . . . . . . . . . . . . . .  17
-4.1.15. SERVER_PORT. . . . . . . . . . . . . . . . . . .  18
4.1.16. SERVER_PROTOCOL. . . . . . . . . . . . . . . . .  18
+4.1.17. SERVER_SOFTWARE. . . . . . . . . . . . . . . . .  19
*/

#if 0
int swebs_set_environment(struct job *job, struct swebs_config *cfg)
{
	char host[64];
	int k;

	job->env = (char**)insistent_zmalloc(NENV*sizeof(char*));
	job->env[NENV-1] = NULL;
	for (k = 0; k < NENV-1; k++)
		job->env[k] = (char*)insistent_zmalloc(ENV_STR_LEN);

	/* AUTH_TYPE 0 */
	job->env[0] = "";
	/* CONTENT_LENGTH 1 */
	job->env[1] = "";
	/* CONTENT_TYPE 2*/
	job->env[2] = "";

	sprintf(job->env[3], "GATEWAY_INTERFACE=%s", "CGI/1.1");

	if (job->hdr.path_info)
		sprintf(job->env[4], "PATH_INFO=%s", job->hdr.path_info);

	/* PATH_TRANSLATED */
	job->env[5] = "";

	/* */
	if (job->hdr.args)
		sprintf(job->env[6], "QUERY_STRING=%s", job->hdr.args);
	else
		job->env[6] = "";

	sprintf(job->env[7], "REMOTE_ADDR=%s", job->who);

	/* REMOTE_HOST */
	job->env[8] = "";

	/* REMOTE_IDENT */
	job->env[9] = "";

	/* REMOTE_USER */
	job->env[10] = "";

	sprintf(job->env[11], "REQUEST_METHOD=%s", job->hdr.method);
	sprintf(job->env[12], "SCRIPT_NAME=%s", job->hdr.file);

	gethostname(host, 63);
	sprintf(job->env[13], "SERVER_NAME=%s", host);

	sprintf(job->env[14], "SERVER_PORT=%d", cfg->tcp_port);
	sprintf(job->env[15], "SERVER_PROTOCOL=%s", job->hdr.http);

	sprintf(job->env[16], "SERVER_SOFTWARE=%s", "swebs");

	/* 4.1.18. Protocol-Specific Meta-Variables . . . . . . . .  19 */
	if (job->hdr.Accept)
		sprintf(job->env[17], "HTTP_ACCEPT=%s", job->hdr.Accept);
	if (job->hdr.User_Agent)
		sprintf(job->env[18], "HTTP_USER_AGENT=%s", job->hdr.User_Agent);
//	if (job->hdr.Referer)
//		sprintf(job->env[19], "HTTP_REFERER=%s", job->hdr.Referer);

	/* set args in case of ISINDEX query */
	if (job->hdr.args) {
		if (!strstr(job->hdr.args, "=")) {
		};
	}
	return 0;
}

int swebs_clear_environment(struct job *job)
{
	int k;
	for (k = 0; k < NENV-1; k++)
		zfree((void**)&job->env[k]);
	zfree((void**)&job->env);
	return 0;
}
#endif

int swebs_send_response(struct job *job)
{
	return http_send_response(
		job->sock, job->status,
		job->when,
		job->content_type,
		job->response_blen,
		job->finfo.st_mtime,
		&job->response_hlen,
		job->from_byte,
		job->upto_byte,
		job->finfo.st_size,
		job->is_last
	);
}

int swebs_send_builtin_response(struct job *job)
{
	int r;
	r = http_send_builtin_response(
				job->sock,
				job->status,
				job->when,
				&job->response_hlen,
				&job->response_blen,
				job->is_last
	);
	if (r > 0)
		job->sent = job->response_blen;
	return r;
}

/*
 * Returns
 *   -1 : read failed, DONT send reply!
 * < -1 : negated http status code
 *    0 : there was no '\r\n\r\n' yet...
 *  > 0 : ok, length of the request
 */

//int swebs_get_request(struct job *job, struct swebs_config *cfg)
int swebs_get_request(struct job *job)
{
	int r;

	//printf("reading from %d\n", job->sock);

	r = recv(job->sock, job->hdr.offs, MAX_HDR_LEN - job->hdr.len, MSG_NOSIGNAL);
	if (r <= 0)
		return -1;

	// printf("## rcvd - \n%s\n", job->hdr.offs);

	job->hdr.len += r;
	job->hdr.offs += r;
	if (job->hdr.len >= MAX_HDR_LEN)
		return -413;

	if (strstr(job->hdr.buf, "\r\n\r\n")) {
		if (job->hdr.len < 16)
			return -400;
		return job->hdr.len;
	};

	return 0;
}

int swebs_parse_http_request(struct job *job)
{
	int err;
//	struct stat finfo;

	err = job->status = http_parse_request_header(
		job->hdr.buf,
		&job->hdr
	);

//	printf("'%s'\n", job->hdr.request_line);
//	printf("+++ meth: '%s'\n", job->hdr.method);
//	printf("+++ uri_: '%s' (err=%d)\n", job->hdr.file, err);
//	printf("+++ http: '%s'\n", job->hdr.http);
//	printf("+++ host: '%s'\n", job->hdr.Host);
//	printf("+++ conn: '%s'\n", job->hdr.Connection);
//	printf("+++ acpt: '%s'\n", job->hdr.Accept);
//	printf("+++ agnt: '%s'\n\n", job->hdr.User_Agent);

	if (err < 0) {
		job->status = -err;
		return err;
	}

	if (job->hdr.file[0] == '.') {
		job->status = 404;
		return -404;
	};

	/* check for DCGI request */
	/* stop further parsing */
	if (job->hdr.method[0] == 'D') {
		// isalpha()?
		// (check corg name!)
		job->status = 600;
		return 0;
	}

	/* check for DCGI path prefix. Here(?) */
	/* swebs_find_cong()....
	if (found) {
		job->cong = ;...
	}
	*/

	/* /index.html */
	if ( !strlen(job->hdr.file) ) {
		strcat(job->index_file, "index.html");
		job->hdr.file = job->index_file;
	};

parse_again:

	err = swebs_parse_uri(job);
	if (err)
		return -err;

	/* /dir/subdir/index.html */

	if (S_ISDIR(job->finfo.st_mode)) {
		strcpy(job->index_file, job->hdr.file);
		if (job->index_file[strlen(job->index_file)-1] != '/')
			strcat(job->index_file, "/");
		strcat(job->index_file, "index.html");
		job->hdr.file = job->index_file;
		goto parse_again;
	};

	if ((job->hdr.Connection) && (toupper(job->hdr.Connection[0]) == 'C')) {
		job->is_last = 1;
	}

//	if ( !(job->finfo.st_mode & S_IXUSR) ) {

//		if (job->hdr.path_info && (strlen(job->hdr.path_info) > 0) ) {
//			job->status = 404;
//			job->hdr.path_info = NULL;
//			return -404;
//		}

		job->response_blen = job->finfo.st_size;
		job->from_byte = job->offs = 0;
		job->upto_byte = job->finfo.st_size - 1;

		http_set_content_type(job->hdr.file, job->content_type);

		if (job->hdr.Range) {

			size_t suflen;
			int r;

			// printf("--- Range: %s\n", job->hdr.Range);

			/* no multiple ranges */
			if (index(job->hdr.Range, ',')) {
				job->status = 416;
				return -416;
			}

			/* first try 'suffix-byte-range-spec' */
			r = sscanf(job->hdr.Range, " bytes = - %d\r\n", &suflen);
			if (1 == r) {
				job->from_byte = job->finfo.st_size - suflen;
				job->upto_byte = job->finfo.st_size - 1;
				goto __range_ok;
			}

			/* now try 'byte-range-spec' */
			r = sscanf(job->hdr.Range, " bytes = %d - %d\r\n", &job->from_byte, &job->upto_byte);
			if (r <= 0) {
				job->status = 416;
				return -416;
			}
			if (!job->upto_byte || job->upto_byte >= job->finfo.st_size)
				job->upto_byte = job->finfo.st_size - 1;

                      __range_ok:
			if (job->upto_byte < job->from_byte || job->from_byte >= job->finfo.st_size) {
				job->status = 416;
				return -416;
			}

			// printf("--- from %d to %d\n", job->from_byte, job->upto_byte);
			job->response_blen = job->upto_byte - job->from_byte + 1;
			job->offs = job->from_byte;
		}
//	};

	return 0;
}

int swebs_parse_uri(struct job *job)
{
	struct http_request_header *h = &job->hdr;
	char *fn = h->file;
	int err;

//	printf("*** swebs_parse_uri(): '%s'\n", h->file);

	/*
	 * TODO: make it understand FULL URIs
	 * (at least just skip "http://host:port/" part if presents)
	 */

	/* skip over 'http://host.domain.dom' if present */
	/*
	scheme = strstr(tok, "http://"); but mind case...
	if (scheme) {
		if (tok != scheme)
			return -400;
		tok += strlen("http://");
		while (tok[0] != '/')
			tok++;
	}
	*/

	if (index(fn, '%')) {
		char *name;
		//curl_easy_unescape();
		name = curl_unescape(fn, 0);
		//printf("--- %s\n", fn);
		//printf("+++ %s\n", name);
		strcpy(fn, name);
		//printf("+++ %s\n", fn);
		curl_free(name);

		name = curl_unescape(job->hdr.request_line, 0);
		strcpy(job->hdr.request_line, name);
		curl_free(name);
	};

	memset(&job->finfo, 0, sizeof(struct stat));
	err = stat(h->file, &job->finfo);
	// printf("*** swebs_parse_uri(): stat '%s': err = %d\n", h->file, err);
	if (-1 == err) {
		job->status = 404;
		return 404;
	};

/*
	do {

		fn = index(fn, '/');
		if (fn)
			fn[0] = 0;

		memset(&job->finfo, 0, sizeof(struct stat));
		err = stat(h->file, &job->finfo);
		// printf("*** swebs_parse_uri(): stat '%s': err = %d\n", h->file, err);
		if (-1 == err) {
			job->status = 404;
			return 404;
		};

		if (!S_ISDIR(job->finfo.st_mode)) {
			if (fn)
				h->path_info = fn + 1;
			break;
		};

		if (!fn)
			break;

		fn[0] = '/';
		fn++;
	} while (1);
*/

/*
	printf("+++ %s\n", job->content_type);
	printf("+++ %s\n", job->hdr.path_info);
	printf("+++ %s\n", job->hdr.args);
*/
	return 0;
}

int swebs_prepare_to_load_file(struct job *job, struct fd_cache *fdc, struct swebs_config *cfg)
{
	size_t buflen = cfg->fragment_size_KiB*1024;
	int fd; 

	fd = fdc_acquire_fd(fdc, job->hdr.file, job->finfo.st_ino, &job->fdci);
	if (-1 == fd)
		return fd;

	if (job->finfo.st_size < buflen)
		buflen = job->finfo.st_size;
	job->buf = (unsigned char*)insistent_zmalloc(buflen);
	job->offs = job->from_byte;

	return fd;
}

int swebs_get_fragment(struct job *job, struct swebs_config *cfg)
{
	size_t buflen = cfg->fragment_size_KiB*1024;
	int r;

	size_t rem = job->response_blen - job->fetched;

	if (buflen > rem)
		buflen = rem;

	if (!buflen) {
		return 0;
	}

	r = insistent_filepread(job->fd, job->buf, buflen, job->offs);
	if (!r) {
		job->is_last = 1;
		return 0;
	}

	job->fetched += r;
	job->offs += r;
	job->buf_sent = 0;
	job->buf_total = r;

//	printf("get_fragment() of '%s': k = %d, len = %d\n", job->hdr.file, k, job->bufs[k].len);

	return r;
}

int swebs_send(struct job *job)
{
	int w;

	w = send(job->sock, job->buf + job->buf_sent, job->buf_total - job->buf_sent, MSG_DONTWAIT | MSG_NOSIGNAL);

	if (w < 0) {
		job->is_last = 1;
		return 1;
	}

	job->buf_sent += w;
	job->sent += w;

	if (job->sent == job->response_blen)
		return 1;

	return 0;
}

void swebs_end_send_file(struct job *job, struct fd_cache *fdc)
{

	if (job->fd)
		fdc_release_fd(fdc, job->hdr.file, job->fd, job->fdci);
	job->fd = -1;
	zfree((void*)&job->buf);
}

#if 0
pid_t swebs_launch_script(struct job *job, int *pipe, struct swebs_config *config, int errlog)
{
	pid_t pid;
	int r;

	while (1) {
		/* pthread_atfork() may be usefull! */
		pid = fork();
		if (-1 == pid) {
			perror("fork()");
			sleep(1);
			continue;
		} else
			break;
	}

	if (pid)
		return pid;

	close(pipe[PIPE_RD_END]);
try_again_1:
	r = dup2(pipe[PIPE_WR_END], STDOUT_FILENO);
	if (-1 == r) {
		LOG_SYS_ERR("dup2", errno);
		sleep(1);
		goto try_again_1;
	}
	// printf("CHLD #### closing %d\n", cgi_pipe[PIPE_RD_END]);

try_again_2:
	//r = dup2(pipe[PIPE_WR_END], STDERR_FILENO);
	r = dup2(errlog, STDERR_FILENO);
	if (-1 == r) {
		LOG_SYS_ERR("dup2", errno);
		sleep(1);
		goto try_again_2;
	}

	swebs_set_environment(job, config);
	// chdir(script dir)!!! spec recommends this
	// close fds: lsock, pipes (except some...), log, etc...!!!
	// or set CLOEXEC или как его там для всех fds
	execve(job->hdr.file, NULL, job->env);
	exit(1);
}
#endif

#if 0
int swebs_read_response_header(struct job *job, struct swebs_config *cfg)
{
	int hdrlen;
//	int totlen;
	int timeo = cfg->cgi_data_timeo;

	hdrlen = http_read_response_header(
				job->fd,
				job->cgihdr.buf,
				sizeof(job->cgihdr.buf)-1, timeo
	);

	// printf("+++ read_hdr(), OK (%d)\n(%s)\n", hdrlen, job->cgihdr.buf);
	if (hdrlen < 0) {
		job->status = -hdrlen;
		return -hdrlen;
	};

	// it's our internal
/*
	if (!strcmp(job->cgihdr.buf, "__YOOPS__: exec() failed.\r\n\r\n")) {
		//printf("+++ OOPS: %d/%d\n///\n%s\n///\n", job->fn_load, hdrlen, job->bufs[k].addr);
		job->status = 500;
		strcpy(job->reason_500, "exec");
		return 500;
	};
*/

	// for now we suppose that was output of syntax correct script, but without http headers...
	// if interpreter returns error, we changes this to 500 [syntax]
/*
	if (!hdrlen && strlen(job->cgihdr.buf)) {
		job->status = 502;
		return 502;
	};
*/
	job->response_hlen = hdrlen;
	return 0;
}
#endif

int swebs_parse_response_header(struct job *job)
{
	return 0;
}

int swebs_record_log(int log, struct job *job)
{
	int err;
	time_t now;
	char timestr_start[32];
	char timestr_now[32];

/* log line format
IP [date1 - date2] REQUEST_LINE (hdr+sent/total) STATUS [reason of 500]
16+        24*2+6 MAX_REQUEST_LINE_LEN+      ~42+     4+             20
*/
	#define LOG_LINE_LEN (16 + 54 + MAX_REQUEST_LINE_LEN + 64 + 4 + 20 + 16)
	char *logline;

	ctime_r(&job->when, timestr_start);
	timestr_start[strlen(timestr_start)-1] = '\0';
	time(&now);
	ctime_r(&now, timestr_now);
	timestr_now[strlen(timestr_now)-1] = '\0';

	logline = (char*)insistent_zmalloc(LOG_LINE_LEN);
	sprintf(
		logline,
		"%s(#%d)\t[%s - %s]\t\"%s\"\t(%d+%d/%d)\t%d",
		job->who,
		job->num,
		timestr_start,
		timestr_now,
		job->hdr.request_line,
		job->response_hlen,
		job->sent,
		job->response_blen,
		job->status
	);

	if (strlen(job->reason_500)) {
		strcat(logline, " [");
		strcat(logline, job->reason_500);
		strcat(logline, "]");
	}
	strcat(logline, "\n");

	flock(log, LOCK_EX);
	err = write(log, logline, strlen(logline));
	flock(log, LOCK_UN);

	free(logline);
	return 0;
}

void swebs_show_stat(struct job_queue *links, int nlinks, struct fd_cache *fdc)
{
	int pn;
	int i = 0;

	while (1) {
		sleep(1);
		//usleep(10000);
		for (pn = 0; pn < nlinks; pn++)
			if (links[pn].load > 0)
				printf("MONITOR: %d jobs %s\n", links[pn].load, links[pn].to_whom);

		/* walk around fd cache sloooowly */
		/* Q: is it really needed? */
		fdc_clear_fd(fdc, i);
		i++; 
		if (i == fdc->nent)
			i = 0;
	};
}

void swebs_generate_addr_mask(unsigned long long *addr)
{
	int k;
	unsigned char rval;
	unsigned long long mask = 0;

	srand((unsigned int)time(NULL));

	for (k = 0; k < 8; k++) {
		rval = (unsigned char)rand();
		mask |= rval;
		if (k < 7)
			mask <<= 8;
	}

	printf("mask = 0x%.16llx\n", mask);
	*addr = mask;
}

#include <crypt.h>

char bread[2048];
char salt[16];

void swebs_gen_jobid(struct job *job, unsigned long long int mask)
{
	char *hash;
	int k;

	memset(bread, 0, sizeof(bread));
	memset(salt, 0, sizeof(salt));

	// add peer TCP port!
	sprintf(bread, "%s:%d#%s#%ld#%p", job->who, 54321, job->hdr.request_line, job->when, job);
	printf("BREAD: '%s'\n", bread);
	sprintf(salt, "$1$%-5s$", job->cong->name);

	if (strlen(job->cong->name) < 5)
		for (k = 3 + strlen(job->cong->name); k < strlen(salt) - 1; k++)
			salt[k] = '_';

	printf("SALT: '%s'\n", salt);

	hash = crypt(bread, salt);
	hash += 3;

	printf("mask: %.16llx, mask^job: %.16llx\n", mask, (unsigned long long int)job ^ mask);
	sprintf(job->id, "%s$%.16llx", hash, (unsigned long long int)job ^ mask);
	/* replace '/' and '.' with чонить другое! */
	printf("JOB-ID: '%s' (mask %.16llx)\n", job->id, mask);
}

//#include <stdio.h>
//#include <stdlib.h>
#include <signal.h>
//#include <unistd.h>
#include <setjmp.h>

void segv_handler(int sig, siginfo_t* sinfo, void * add_info);
struct sigaction segv_action;
struct sigaction segv_action_old;
// sigset_t blocked;

jmp_buf jump_data;
volatile int bad_addr;

struct job *swebs_check_jobid(char *jobid, unsigned long long int mask)
{
	char *bux;
	char *id;
//	char *salt;
	char *jaddr;

	struct job *job;
	unsigned long long int addr;

	int r, match;

	/* parse X-DCGP-JOB-ID from a CONG */

	printf("cheking JOB-ID '%s'\n", jobid);

	bux = index(jobid, '$');
	if (!bux) {
		printf("no '$' #1\n");
		return NULL;
	}

//	bux[0] = 0;
//	salt = jobid;

	id = ++bux;
	bux = index(id, '$');

	if (!bux) {
		printf("no '$' #2\n");
		return NULL;
	}

//	bux[0] = 0;
	jaddr = ++bux;

	r = sscanf(jaddr, "%llx", &addr);
	if (1 != r) {
		printf("sscanf() failed\n");
		return NULL;
	}

	addr ^= mask;
	job = (struct job*)((unsigned)addr);
	printf("Checking addr %p... \n", job);

	// set SIGSEGV handler
	segv_action.sa_handler = NULL;
	segv_action.sa_sigaction = segv_handler;
	segv_action.sa_flags = SA_SIGINFO;
	segv_action.sa_restorer = NULL;

	r = sigaction(SIGSEGV, &segv_action, &segv_action_old);
	if (-1 == r) {
		perror("[1] sigaction()");
		exit(1);
	};

	bad_addr = 0;
	setjmp(jump_data);

	/* now try accessing memory using the addr */
	if (bad_addr)
		return NULL;
	else {
		printf("trying to compare '%s' and '%s'... \n", job->id, jobid);
		match = !strcmp(job->id, jobid);
	}

	printf("addr OK, checking ID ...\n");

	r = sigaction(SIGSEGV, &segv_action_old, NULL);
	if (-1 == r) {
		perror("[2] sigaction()");
		exit(1);
	};

	if (!match)
		return NULL;

	printf("OK\n");
	/* if we get here, everything is all right */
	return job;
}

void segv_handler(int sig, siginfo_t *sinfo, void *add_info)
{

/*
	printf("\nSignal info:\n~~~~~~~~~~~~\n");
	printf("numb:\t %d (%d)\n", sinfo->si_signo, sig);
	printf("errn:\t %d\n", sinfo->si_errno);
	printf("code:\t %d ", sinfo->si_code);

	switch (sinfo->si_code) {

		case SEGV_MAPERR:
			printf("(SEGV_MAPERR - \
			address not mapped to object)\n");
		break;

		case SEGV_ACCERR:
			printf("(SEGV_ACCERR - \
			invalid permissions for mapped object)\n");
		break;

		default:
			printf("???\n");
	};
*/

	printf("SIGSEGV HANDLER, faulty addr:\t 0x%.8X\n", (unsigned int)sinfo->si_addr);
	bad_addr = 1;
	longjmp(jump_data, 0);
}
